/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */

#ifndef QUB_IDLSEG_H
#define QUB_IDLSEG_H

/*
This module handles in-memory storage of "Idealizations" of sampled data.

An Idealization assigns each sample to one of a limited number of Classes,
then condenses this sequence of sampled Classes into a sequence of Dwells:
{ which class; start sample index; end sample index (inclusive) }
Classes are numbered contiguously starting with 0.
This module keeps track of each Class's {mean amplitude; standard deviation}.

Previously, Idealizations had been stored as parallel arrays:
int class[]
int first[]
int last[]
(and/or float duration[])

There were problems with this approach:
 * editing a portion could change the number of dwells,
   requiring shifts and/or reallocation
 * to read the dwells for samples k through m,
   we had to search for the first dwell,
   since dwell index doesn't correlate with sample index
 * the arrays are all in memory,
   which for a cd-full of busy data could be ~400MB

To address these, we use advanced data structures with log(N) edits and seeks
(at the core, we use the STL map),
and move "chunks" of idealization between three storage levels:
 * in STL maps: efficient editing and seeking; biggest memory footprint
 * in arrays:   no editing beyond full replacement; smaller memory footprint
 * paged to a temp file as arrays

This module also provides:
 * support for segmentation
      A sampled file might be segmented if it contains only portions of
	  the original recording.  This is relevant to Idealizations because
	  a Dwell should not cross segment boundaries.  In addition, qub allows
	  users to modify segmentation e.g. by breaking a segment in two or cut/paste,
	  and the idealization must follow these changes.

	  Somewhat confusingly, a data segment might be represented in a
	  DataSets qubtree by several idealized segments, if those are separated
	  by un-idealized intervals.
 * import/export from other formats
    - qub DWT (read-only)
	- axon EVL (read-only)
	  (and it's confusing because the record sometimes has gaps ?!)
    - first/last/class arrays (read-only; as a way of overwriting some or all events)
	- DataSet qubtree containing first/last/class or duration/class arrays
	  (read/write)
	  (http://www.qub.buffalo.edu/qubdoc/trees/DataSets.html)
    - ASCII (write-only)
	  writes the class number of each sample
 * thread-safety
      The Idealization object has a critical section which is managed
	  automatically by all methods and iterators, and you can
	  lock()/unlock() for longer periods of time e.g. for non-atomic edits

More about the data structures:

The idealization is divided into segments, according to the sampled data segmentation.
Each segment is divided into chunks of QUB_IDLSEG_CHUNKSIZE samples:
  - for lookup: dwells for sample k are in chunk (k >> QUB_IDLSEG_CHUNKSHIFT)
  - for map traversal and editing: N events take N log N time,
    but since there are only so many events per map,
	it becomes N log M, where log M has a constant upper bound
  - for multiple storage levels: only the relevant chunks need to be loaded
    for reading, or upgraded to "map" for editing.
The chunk size has been experimentally tuned to handle fast data smoothly.

A Segment has an IdlMaps object which manages one or more chunks.

Every sample is said to belong to a class.  Un-idealized samples belong to dwells
of class -1.

You shouldn't have to work directly with chunks or segments.  Instead,
use IdlSegments and IdlSegments_Iter.  Here are the highlights:

QUB_IdlSegments(double sampling);  // sampling is the time in seconds between two samples
~QUB_IdlSegments();
void __cdecl Free();               // required for Delphi compatibility
                                   // as are all _whatever() methods

// to enable paging-out-to-disk, call useTempFile(1)
int        __cdecl useTempFile(int doUse);

// saves the idealization as a sampled textfile, a stream of class numbers
int        __cdecl saveASCII(char *fileName);

void       __cdecl lock();
void       __cdecl unlock();
// all methods and iterators handle locking automatically;
// lock() and unlock() yourself only to thread-protect multiple operations
// that need to be atomic.

double     __cdecl getSampling();            // in seconds
void       __cdecl setSampling(double samp);
int        __cdecl getClsCount();            // number of Classes
int        __cdecl getSegCount();            // number of Segments

QUB_Tree   __cdecl getSegmentation();
// returns a qub tree with int data[segcount][2] containing
// the first and last sample indices of each segment

QUB_Tree   __cdecl toTree();
// returns the full idealization as a DataSet tree
// (http://www.qub.buffalo.edu/qubdoc/trees/DataSets.html)

int        __cdecl readDWT(char *fileName);
// DWT file (http://www.qub.buffalo.edu/qubdoc/files/dwt.html)

// routines for reading a portion of the idealization:
// options:
//    * f, l:       the first and last sample indices
//    * fragments:  if the [f,l] interval begins/ends in the middle of a dwell,
//                  do you want the partial dwell(s)?
//    * withDurs:   it's a waste of time and space to calculate the duration
//                  of each event if you're just interested in the first/last sample
//    * dataSet:        DataSet qubtree to add idealized Segments to
//    * insertAsFirst:  it's a lot quicker to insert a subtree as the first child
//                      than to append as the last.  With care, this can save a
//                      bit of time, but is probably overkill in most situations.
//    * dropFirstLast:  leave out the first and last events in each idl segment
//    * dropIfCls:      drop the first and/or last event only if they match dropCls
// getDwells():          returns a DataSet qubtree with one or more Segment children
// getDwellsInto():      inserts or appends one or more Segment children to an
//                       existing DataSet qubtree
// getDwellsIntoSegs():  fills an existing DataSet qubtree's existing Segment children
//                       with idealized data, according to the Segments' bounds
// getDwellsIntoWDrop(): like getDwellsInto(), with the option to drop first/last events
QUB_Tree   __cdecl getDwells(int f, int l, int fragments, int withDurs);
void       __cdecl getDwellsInto(int f, int l, int fragments, int withDurs,
								QUB_Tree dataSet, int insertAsFirst );
void       __cdecl getDwellsIntoSegs(QUB_Tree dataSet, int withDurs);
void       __cdecl getDwellsIntoWDrop(int f, int l, int withDurs,
								int dropFirstLast, int dropIfCls, int dropCls,
								QUB_Tree dataSet, int insertAsFirst );

// routines for modifying the idealization:
// setDwells():    reads from Segment children of a DataSet qubtree
// setDwellsSeg(): reads from a Segment qubtree
// setDwells():    reads from an array of {cls,first,last} structs
// setDwellsFLC(): reads from parallel arrays of first[] last[] class[]
int        __cdecl setDwells(QUB_Tree dataSet);
int        __cdecl setDwellsSeg(QUB_Tree seg);
int        __cdecl setDwells(int count, QUB_IdlDwell *buf);
int        __cdecl setDwellsFLC(int count, int *ff, int *ll, int *cc);

// clear():   replaces the whole thing with one long dwell of class -1
int        __cdecl clear();

// erase():   replaces the idealization in the selection [f,l] with class -1
int        __cdecl erase(int f, int l);
// join():    extends the dwell at sample f so it extends through sample l
//            e.g. to replace a false opening with whatever class preceded it
int        __cdecl join(int f, int l);

// routines for modifying data segmentation:
// addSeg(): do this initially for each data segment
int        __cdecl addSeg(int f, int l);
// removePoints(): e.g. when the user deletes a selection of data
int        __cdecl removePoints(int f, int l);
// insertPoints(): e.g. when the user pastes in a selection of data
int        __cdecl insertPoints(int f, int l);
// breakSeg(): most commonly part of a behind-the-scenes sequence;
// e.g. the user deletes a selection of data but you don't want to
// join the edges together, so you removePoints() then breakSeg()
int        __cdecl breakSeg(int at);
// joinSegs(): only really used to undo breakSeg()
int        __cdecl joinSegs(int at);

// classic c++ style iteration:
typedef QUB_IdlSegmentsIter iterator;
iterator           begin();
iterator           end();
iterator           find(int point);

// These iterators are a little different from what you see in
// the first,last,class arrays:
//    * in non-idealized regions, the iterator points to dwells of class -1
//    * Dwell bounds (first, last) are offset by their segment's "first" sample index
//    * two adjacent dwells may have the same class
// Recall that segments are broken into chunks.  A side-effect is that a
// particularly long dwell is broken into sub-dwells.
// This style is appropriate for display (less processing -> faster),
// but you can spare the effort of re-joining dwells, adding seg-first-index,
// and detecting idl-segment bounds (non-idealized regions)
// by using getDwells*() instead.

// Here are the methods of the iterator:

// assignment and equality testing:
QUB_IdlSegmentsIter(const QUB_IdlSegmentsIter &other)
QUB_IdlSegmentsIter& operator= (const QUB_IdlSegmentsIter& other)
bool operator==(const QUB_IdlSegmentsIter& other)

// validity testing (boolean cast is true unless the iterator == idealization.end())
operator bool ()

// dereferencing (looking at the current Dwell):
QUB_IdlDwell& operator* () const { return *iter; }
QUB_IdlDwell* operator->() const { return &(*iter); }

// iter->first, last are in seg.local terms --
// add segfirst() to find location in whole data file
int segfirst() const { return seg->first; }
int segindex() const { return seg_ix; }

// read() copies all the info from *iter,
// adds segfirst() to f and l, and looks up the mean amp of class c:
void __cdecl read( int *f, int *l, int *c, double *a );

// next/prev dwell:
QUB_IdlSegmentsIter& operator++ ();
QUB_IdlSegmentsIter& operator-- ();

// Delphi compatibility (duplicates of c++ operator overloads):
void __cdecl Free();
int __cdecl valid();
void __cdecl next();
void __cdecl prev();
QUB_IdlSegmentsIter_iface * __cdecl copy();



*/

#include "QUB_Freezer.h"
#include "QUB_NodeArray.h"

#define QUB_IDLSEG_CHUNKSHIFT 14
#define QUB_IDLSEG_CHUNKSIZE (1 << QUB_IDLSEG_CHUNKSHIFT)

typedef struct
{
	int cls;
	int first;
	int last;
} QUB_IdlDwell;


typedef std::map< int, QUB_IdlDwell >  QUB_IdlMap;


class QUB_IdlChunk
{
protected:
	void setComplete(int count, QUB_IdlDwell *buf, int offset, int pointsInSeg);
	void setComplete(int count, int *ff, int *ll, int *cc, int offset, int pointsInSeg);

public:
	int first;
	QUB_NodeStash& stash;

	int storage;
	QUB_IdlMap *asMap;

	int stashKey;
	QUB_Tree asNode;

	QUB_IdlChunk(int afirst, QUB_NodeStash& storage);
	virtual ~QUB_IdlChunk();

	int size();
	void setupStorage(int lvl);
	void clearStorage(int lvl);
	int  getStorage();

	// the rest of the methods do nothing unless you have storage level 0
	void clear(int f, int l);
	void set(int f, int l, int c, bool atBegin, bool atEnd); // assumes area is cleared

	// clears affected, returns offset in buf of first dwell not completely consumed
	int set(int count, QUB_IdlDwell *buf, int offset, int pointsInSeg); // offset == seg.first
	int set(int count, int *ff, int *ll, int *cc, int offset, int pointsInSeg);
};

typedef QUB_Freezer<QUB_IdlChunk> QUB_IdlFreezer;
typedef QUB_FreezerSeries<QUB_IdlChunk> QUB_IdlFreezerSeries;


class QUB_IdlMaps;

class QUB_IdlMapsIter
{
private:
	QUB_IdlMaps* maps;

public:
	QUB_IdlChunk *chunk;
	int          ix;
	
	QUB_IdlMap::iterator iter;

	QUB_IdlDwell *arr;
	int          i, count;

	QUB_IdlMapsIter()
		: maps(0), chunk(0), ix(-1), arr(0)
	{}
	QUB_IdlMapsIter(const QUB_IdlMapsIter& other)
		: maps(0), chunk(0), ix(-1), arr(0)
	{
		*this = other;
	}

	QUB_IdlMapsIter(QUB_IdlMaps& amaps, int point );
	QUB_IdlMapsIter(QUB_IdlMaps& amaps, bool atBegin);
	virtual ~QUB_IdlMapsIter();

	QUB_IdlMapsIter& operator=(const QUB_IdlMapsIter& other);

	bool operator==(const QUB_IdlMapsIter& other) const { return arr ? (arr == other.arr && i == other.i) : iter == other.iter; }
	bool operator!=(const QUB_IdlMapsIter& other) const { return ! (*this == other); }
	operator bool () const { return chunk && (arr ? (i != count) : (iter != chunk->asMap->end())); }
	QUB_IdlDwell& operator* () const { return arr ? arr[i] : iter->second; }
	QUB_IdlDwell* operator->() const { return &(arr ? arr[i] : iter->second); }

	QUB_IdlMapsIter& operator++ ();
	QUB_IdlMapsIter& operator-- ();
};



class QUB_IdlMaps
{
private:
	void setSize( int n );
public:

	QUB_IdlFreezer& freezer;
	QUB_IdlFreezerSeries chunks;
	QUB_NodeStash& chunkStorage;
	int maxmap;
	int points;

	QUB_IdlMaps(QUB_IdlFreezer& freezer, QUB_NodeStash& stash, int points);
	virtual ~QUB_IdlMaps();

	void insert(int f, int l);
	void remove(int f, int l);
	
	void set(int f, int l, int c);
	void join(int f, int l);

	// return offset in buf of first dwell not completely consumed
	int set(int count, QUB_IdlDwell *buf, int offset); // offset == seg.first
	int set(int count, int *ff, int *ll, int *cc, int offset);

	typedef QUB_IdlMapsIter iterator;

	iterator begin()         { return QUB_IdlMapsIter( *this, true ); }  // { return maps.size() ? iterator(maps, maxmap, 0, maps[0].begin()) : end(); }
	iterator end()           { return QUB_IdlMapsIter( *this, false ); } // { return iterator(maps, maxmap, maxmap, (maxmap+1) ? maps[maxmap].end() : QUB_IdlMap::iterator()); }
	iterator find(int point) { return QUB_IdlMapsIter( *this, point ); }
};

typedef std::vector< std::pair<float, float> > QUB_IdlAmpStd;

class QUB_IdlSeg
{
public:
	QUB_IdlFreezer&      freezer;
	QUB_NodeStash&       chunkStorage;
	double               sampling;
	int                  first;
	int                  last;
	QUB_IdlMaps          maps;
	QUB_IdlAmpStd        ampStd;


	QUB_IdlSeg(QUB_IdlFreezer& freezer, QUB_NodeStash& stash, int first, int last, double samp);

	void      setFirst(int f);
	void      setLast(int l);
	void      offset(int points);

	void      insert(int f, int l);
	void      remove(int f, int l);
	
	void      set(int f, int l, int c);
	void      join(int f, int l);

	void      getDwells(int f, int l, int fragments, int withDurs, QUB_Tree dataSet, QUB_Tree& afterSeg );
	void      getDwellsWDrop(int f, int l, int withDurs, int dropFL, int dropIfCls, int dropCls, QUB_Tree dataSet, QUB_Tree& afterSeg );
	void      getDwellsInto(QUB_Tree segment, int withDurs);
	void      setDwells(QUB_Tree segment);
	void      setDwells(int count, QUB_IdlDwell *buf);
	void      setDwells(int count, int *ff, int *ll, int *cc);

	int       getClassCount();
	void      setClassCount(int cc);
	double    getAmp(int cls);
	void      setAmp(int cls, double amp);
	double    getStd(int cls);
	void      setStd(int cls, double std);

	QUB_IdlMaps::iterator begin()         { return maps.begin(); }
	QUB_IdlMaps::iterator end()           { return maps.end(); }
	QUB_IdlMaps::iterator find(int point) { return maps.find(point - first); }
};




class QUB_IdlSegmentsIter_iface
{
public:
	virtual void __cdecl Free() = 0;

	virtual int __cdecl valid() = 0;
	virtual void __cdecl read( int *f, int *l, int *c, double *a ) = 0;
	virtual void __cdecl next() = 0;
	virtual void __cdecl prev() = 0;

	virtual QUB_IdlSegmentsIter_iface * __cdecl copy() = 0;
};



class QUB_IdlSegments_iface
{
public:
	virtual void __cdecl Free() = 0;

	virtual void       __cdecl lock() = 0;
	virtual void       __cdecl unlock() = 0;

	virtual double     __cdecl getSampling() = 0;
	virtual void       __cdecl setSampling(double samp) = 0;
	virtual int        __cdecl getClsCount() = 0;
	virtual int        __cdecl getSegCount() = 0;
	virtual QTR_Impl * __cdecl _getSegmentation() = 0;

	virtual QTR_Impl * __cdecl _toTree() = 0;
	virtual int        __cdecl readDWT(char *fileName) = 0;
	virtual QTR_Impl * __cdecl _getDwells(int f, int l, int fragments, int withDurs) = 0;
	virtual void       __cdecl _getDwellsInto(int f, int l, int fragments, int withDurs,
										  QTR_Impl *dataSet, int insertAsFirst ) = 0;
	virtual void       __cdecl _getDwellsIntoSegs(QTR_Impl *dataSet, int withDurs) = 0;
	virtual void       __cdecl _getDwellsIntoWDrop(int f, int l, int withDurs,
								  int dropFirstLast, int dropIfCls, int dropCls,
								  QTR_Impl *dataSet, int insertAsFirst ) = 0;

	virtual int        __cdecl _setDwells(QTR_Impl *dataSet) = 0;
	virtual int        __cdecl _setDwellsSeg(QTR_Impl *seg) = 0;

	virtual int        __cdecl setDwells(int count, QUB_IdlDwell *buf) = 0;
	virtual int        __cdecl setDwellsFLC(int count, int *ff, int *ll, int *cc) = 0;

	virtual int        __cdecl clear() = 0;
	virtual int        __cdecl addSeg(int f, int l) = 0;
	virtual int        __cdecl removePoints(int f, int l) = 0;
	virtual int        __cdecl insertPoints(int f, int l) = 0;
	virtual int        __cdecl breakSeg(int at) = 0;
	virtual int        __cdecl joinSegs(int at) = 0;

	virtual int        __cdecl erase(int f, int l) = 0;
	virtual int        __cdecl join(int f, int l) = 0;

	virtual QUB_IdlSegmentsIter_iface* __cdecl _begin() = 0;
	virtual QUB_IdlSegmentsIter_iface* __cdecl _end() = 0;
	virtual QUB_IdlSegmentsIter_iface* __cdecl _find(int point) = 0;

	virtual int        __cdecl useTempFile(int doUse) = 0;

	virtual int        __cdecl saveASCII(char *fileName) = 0;
};


class QUB_IdlSegmentsIter;

class QUB_IdlSegments : public QUB_IdlSegments_iface
{
private:
	QUB_IdlFreezer      freezer;
	QUB_NodeStash       chunkStorage;
	CRITICAL_SECTION    crit;

	void       getSegRange(int f, int l, int& sf, int& sl);
	void       resetSeg(int s);

public:
	std::vector<QUB_IdlSeg*> segs;
	double              sampling;

	QUB_IdlSegments(double sampling);
	~QUB_IdlSegments();
	void __cdecl Free();

	void       __cdecl lock();
	void       __cdecl unlock();

	double     __cdecl getSampling();
	void       __cdecl setSampling(double samp);
	int        __cdecl getClsCount();
	int        __cdecl getSegCount();
	QTR_Impl * __cdecl _getSegmentation();
	QUB_Tree   __cdecl getSegmentation();

	QTR_Impl * __cdecl _toTree();
	QUB_Tree   __cdecl toTree();
	int        __cdecl readDWT(char *fileName);
	QTR_Impl * __cdecl _getDwells(int f, int l, int fragments, int withDurs);
	QUB_Tree   __cdecl getDwells(int f, int l, int fragments, int withDurs);
	void       __cdecl _getDwellsInto(int f, int l, int fragments, int withDurs,
								  QTR_Impl *dataSet, int insertAsFirst );
	void       __cdecl getDwellsInto(int f, int l, int fragments, int withDurs,
								 QUB_Tree dataSet, int insertAsFirst );
	void       __cdecl _getDwellsIntoSegs(QTR_Impl *dataSet, int withDurs);
	void       __cdecl getDwellsIntoSegs(QUB_Tree dataSet, int withDurs);
	void       __cdecl _getDwellsIntoWDrop(int f, int l, int withDurs,
								  int dropFirstLast, int dropIfCls, int dropCls,
								  QTR_Impl *dataSet, int insertAsFirst );
	void       __cdecl getDwellsIntoWDrop(int f, int l, int withDurs,
								  int dropFirstLast, int dropIfCls, int dropCls,
								  QUB_Tree dataSet, int insertAsFirst );
	// requires no "Segment" crossing idl segment bounds

	int        __cdecl _setDwells(QTR_Impl *dataSet);
	int        __cdecl setDwells(QUB_Tree dataSet);
	int        __cdecl _setDwellsSeg(QTR_Impl *seg);
	int        __cdecl setDwellsSeg(QUB_Tree seg);
	int        __cdecl setDwells(int count, QUB_IdlDwell *buf);
	int        __cdecl setDwellsFLC(int count, int *ff, int *ll, int *cc);

	int        __cdecl clear();
	int        __cdecl addSeg(int f, int l);
	int        __cdecl removePoints(int f, int l);
	int        __cdecl insertPoints(int f, int l);
	int        __cdecl breakSeg(int at);
	int        __cdecl joinSegs(int at);

	int        __cdecl erase(int f, int l);
	int        __cdecl join(int f, int l);

	typedef QUB_IdlSegmentsIter iterator;
	iterator           begin();
	iterator           end();
	iterator           find(int point);

	QUB_IdlSegmentsIter_iface* __cdecl _begin();
	QUB_IdlSegmentsIter_iface* __cdecl _end();
	QUB_IdlSegmentsIter_iface* __cdecl _find(int point);

	int        __cdecl useTempFile(int doUse);

	int        __cdecl saveASCII(char *fileName);
};


class QUB_IdlSegmentsIter : public QUB_IdlSegmentsIter_iface
{
private:
	QUB_IdlSegments&       segs;
	int                    seg_ix;
	QUB_IdlMaps::iterator  iter;
	QUB_IdlSeg*            seg;

public:
	QUB_IdlSegmentsIter(QUB_IdlSegments& asegs, int aseg_ix, QUB_IdlMaps::iterator& aiter )
		: segs( asegs ), seg_ix( aseg_ix ), iter( aiter ), seg( (seg_ix>=0) ? segs.segs[seg_ix] : NULL )
	{
		segs.lock();
	}

	QUB_IdlSegmentsIter(const QUB_IdlSegmentsIter &other)
		: segs( other.segs ), seg_ix( other.seg_ix ), iter( other.iter ), seg( other.seg )
	{
		segs.lock();
	}

	virtual ~QUB_IdlSegmentsIter()
	{
		segs.unlock();
	}

	QUB_IdlSegmentsIter& operator= (const QUB_IdlSegmentsIter& other)
	{
		throw "undefined";
	}

	bool operator==(const QUB_IdlSegmentsIter& other) const { return iter == other.iter; }
	operator bool () const { return seg && iter; } // (iter != seg->end()); }
	QUB_IdlDwell& operator* () const { return *iter; }
	QUB_IdlDwell* operator->() const { return &(*iter); }
	// iter->first, last are in seg.local terms -- add segfirst() to find location in whole data file
	int segfirst() const { return seg->first; }
	int segindex() const { return seg_ix; }

	QUB_IdlSegmentsIter& operator++ ();
	QUB_IdlSegmentsIter& operator-- ();

	void __cdecl Free();

	int __cdecl valid();
	void __cdecl read( int *f, int *l, int *c, double *a );
	void __cdecl next();
	void __cdecl prev();

	QUB_IdlSegmentsIter_iface * __cdecl copy();
};


extern "C" MIL_API QUB_IdlSegments_iface *    __stdcall QUB_IdlSegments_Create( double sampling );


extern "C" MIL_API QTR_Impl * __stdcall QUB_Idl_ReadDWT(char *filename);
extern "C" MIL_API QTR_Impl * __stdcall QUB_Idl_ReadEVL(char *filename);
extern "C" MIL_API QTR_Impl * __stdcall QUB_Idl_ReadTAC(char *filename);


#endif
